@@ -1,6 +1,8 @@
 import json
 import logging
 import traceback as tb
+from dataclasses import dataclass
+from typing import NewType, Optional
 from daggerml._config import DML_API_ENDPOINT, DML_API_KEY
 from http.client import HTTPConnection, HTTPSConnection
 from urllib.parse import urlparse
@@ -35,12 +37,31 @@ class NodeError(DmlError):
     pass
 
 
+@dataclass(frozen=True)
+class Resource:
+    id: str
+    parent: Optional[NewType('Resource', None)]
+    tag: Optional[str] = None
+
+    @classmethod
+    def from_dict(cls, data):
+        if data['parent'] is None:
+            return cls(data['id'], data['tag'], None)
+        return cls(data['id'], cls.from_dict(data['parent']), data['tag'])
+
+    def to_dict(self):
+        parent = None
+        if self.parent is not None:
+            parent = self.parent.to_dict()
+        return {'id': self.id, 'tag': self.tag, 'parent': parent}
+
+
 def _api(api, op, group=None, **kwargs):
     try:
         url = urlparse(DML_API_ENDPOINT)
         scheme = url.scheme
         host = url.hostname
-        port = url.port
+        port = url.port or 443
         path = url.path
         assert all(x is not None for x in [scheme, host, port, path]), \
             f'invalid endpoint URL: {DML_API_ENDPOINT}'
@@ -77,11 +98,16 @@ def describe_dag(dag_id):
     return _api('dag', 'describe', dag_id=dag_id)
 
 
+def get_dag_by_name_version(dag_name, version='latest'):
+    tmp = _api('dag', 'get_dag_by_name_version', name=dag_name, version=version)
+    if tmp is None:
+        return None
+    return tmp['result']
+
+
 def format_exception(err):
     if isinstance(err, NodeError):
-        return {
-            'message': err.msg['message']
-        }
+        return err.msg
     return {
         'message': str(err),
         'trace': tb.format_exception(type(err), value=err, tb=err.__traceback__)
@@ -92,25 +118,14 @@ def daggerml():
     from time import sleep
     from collections.abc import Mapping
     from weakref import WeakKeyDictionary
-    from dataclasses import dataclass
-    from typing import NewType, Optional
     from uuid import uuid4
 
-    @dataclass(frozen=True)
-    class Resource:
-        id: str
-        parent: Optional[NewType('Resource', None)]
+    tag2resource = {}
 
-        @classmethod
-        def from_json(cls, data):
-            if data['parent'] is None:
-                return cls(data['id'], None)
-            return cls(data['id'], cls.from_json(data['parent']))
-
-        def to_json(self):
-            if self.parent is None:
-                return {'id': self.id, 'parent': None}
-            return {'id': self.id, 'parent': self.parent.to_json()}
+    def register_tag(tag, cls):
+        assert tag not in tag2resource
+        tag2resource[tag] = cls
+        return
 
     def to_data(py, dag=None):
         if isinstance(py, Node):
@@ -140,7 +155,7 @@ def daggerml():
         elif isinstance(py, float):
             return {'type': 'scalar', 'value': {'type': 'float', 'value': str(py)}}
         elif isinstance(py, Resource):
-            return {'type': 'resource', 'value': py.to_json()}
+            return {'type': 'resource', 'value': py.to_dict()}
         else:
             raise ValueError('unknown type: ' + type(py))
 
@@ -165,7 +180,7 @@ def daggerml():
             else:
                 raise ValueError('unknown scalar type: ' + t)
         elif t == 'resource':
-            return Resource.from_json(v)
+            return tag2resource.get(v['tag'], Resource).from_dict(v)
         else:
             raise ValueError('unknown type: ' + t)
 
@@ -261,12 +276,6 @@ def daggerml():
                 self.check()
             return self.result
 
-    def get_dag_by_name_version(dag_name, version='latest'):
-        tmp = _api('dag', 'get_dag_by_name_version', name=dag_name, version=version)
-        if tmp is not None:
-            tmp = tmp['result']
-        return tmp
-
     @dataclass(frozen=True)
     class Dag:
         id: str
@@ -278,16 +287,27 @@ def daggerml():
         executor_id: str = None
         secret: str = None
 
-        def __post_init__(self):
-            if self.executor_id is not None:
-                object.__setattr__(self, 'executor', Node(self, self.executor_id).to_py())
-            pass
-
         @classmethod
         def new(cls, name, group='test0'):
             resp = _api('dag', 'create_dag', name=name, group=group)
             return cls(**resp, group=group)
 
+        @classmethod
+        def from_claim(cls, executor, secret, ttl, group, node_id=None):
+            resp = _api('node', 'claim_node', executor=executor.to_dict(),
+                        ttl=ttl, node_id=node_id, group=group, secret=secret)
+            if resp is None:
+                return
+            return cls(**resp, group=group)
+
+        @property
+        def expr(self):
+            return Node(self, self.expr_id)
+
+        @property
+        def executor(self):
+            return Node(self, self.executor_id).to_py()
+
         def from_py(self, py):
             if isinstance(py, Node):
                 return py
@@ -319,6 +339,14 @@ def daggerml():
                  group=self.group, secret=self.secret)
             return
 
+        def refresh(self, ttl=300):
+            res = _api('dag', 'refresh_claim', dag_id=self.id,
+                       group=self.group, secret=self.secret, ttl=ttl,
+                       refresh_token=self.id)
+            if res is None:
+                raise DagError('Failed to refresh dag!')
+            return
+
         def load(self, dag_name, version='latest'):
             node_id = get_dag_by_name_version(dag_name, version)
             if node_id is None:
@@ -327,9 +355,9 @@ def daggerml():
                        node_id=node_id, secret=self.secret)
             return Node(self, res['node_id'])
 
-        def create_resource(self):
+        def create_resource(self, tag=None):
             res = _api('dag', 'create_resource', dag_id=self.id,
-                       group=self.group, secret=self.secret)
+                       group=self.group, secret=self.secret, tag=tag)
             return Node(self, res['node_id']), res['secret']
 
         def __repr__(self):
@@ -343,14 +371,8 @@ def daggerml():
                 self.fail(format_exception(exc_val))
                 return True  # FIXME remove this to not catch these errors
 
-    def claim_execution(executor, secret, ttl, node_id=None, group='test0'):
-        resp = _api('node', 'claim_node', executor=executor.to_json(),
-                    ttl=ttl, node_id=node_id, group=group, secret=secret)
-        resp['group'] = group
-        return resp
-
-    return Resource, Dag, Node, claim_execution
+    return Dag, Node, register_tag
 
 
-Resource, Dag, Node, claim_execution = daggerml()
+Dag, Node, register_tag = daggerml()
 del daggerml
